"""
Merge duration and tempo in routine_activity.

Revision ID: 06f82ead211b
Revises: 22f3ddb25741
Create Date: 2023-04-02

"""

import sqlalchemy as sa
from alembic import op

revision = "06f82ead211b"
down_revision = "22f3ddb25741"
branch_labels = None
depends_on = None


check_constraints_up = [
    ("time_type_integer", "typeof(time) = 'integer'"),
    ("time_ge_0", "time >= 0"),
]
check_constraints_down = [
    ("duration_type_integer", "typeof(duration) = 'integer'"),
    ("tempo_type_integer", "typeof(tempo) = 'integer'"),
    ("duration_ge_0", "duration >= 0"),
    ("tempo_ge_0", "tempo >= 0"),
]


def upgrade() -> None:
    with op.batch_alter_table("routine_activity", schema=None) as batch_op:
        for constraint_name, _ in check_constraints_down:
            batch_op.drop_constraint(constraint_name, type_="check")

        batch_op.alter_column("duration", new_column_name="time")
        batch_op.drop_column("tempo")

        for constraint_name, condition in check_constraints_up:
            batch_op.create_check_constraint(constraint_name, condition)


def downgrade() -> None:
    with op.batch_alter_table("routine_activity", schema=None) as batch_op:
        for constraint_name, _ in check_constraints_up:
            batch_op.drop_constraint(constraint_name, type_="check")

        batch_op.alter_column("time", new_column_name="duration")
        batch_op.add_column(sa.Column("tempo", sa.INTEGER(), nullable=False, default=0))

        for constraint_name, condition in check_constraints_down:
            batch_op.create_check_constraint(constraint_name, condition)



from logging.config import fileConfig

from alembic import context

from valens import app, database as db, models

config = context.config

if config.config_file_name:
    fileConfig(config.config_file_name)

TARGET_METADATA = models.Base.metadata

assert not context.is_offline_mode()

with app.app_context():
    # Prevent integrity errors and data loss caused by ON DELETE cascades
    app.config["SQLITE_FOREIGN_KEY_SUPPORT"] = False

    connectable = context.config.attributes.get("connection", None)

    if connectable is None:
        connectable = db.get_engine()

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=TARGET_METADATA,
            render_as_batch=True,
        )

        with context.begin_transaction():
            context.run_migrations()

    app.config["SQLITE_FOREIGN_KEY_SUPPORT"] = True



"""
Extend workouts.

Revision ID: 8a0dc258bf2a
Revises: b9f4e42c7135
Create Date: 2023-02-16

"""

from typing import Union

import sqlalchemy as sa
from alembic import op

revision = "8a0dc258bf2a"
down_revision = "b9f4e42c7135"
branch_labels = None
depends_on = None


workout_set = sa.table(
    "workout_set",
    sa.column("workout_id", sa.Integer),
    sa.column("position", sa.Integer),
    sa.column("exercise_id", sa.Integer),
    sa.column("reps", sa.Integer),
    sa.column("time", sa.Integer),
    sa.column("weight", sa.Float),
    sa.column("rpe", sa.Float),
)
workout_element = sa.table(
    "workout_element",
    sa.column("workout_id", sa.Integer),
    sa.column("position", sa.Integer),
    sa.column("type", sa.Text),
    sa.column("automatic", sa.Boolean),
)

check_constraints: list[tuple[str, Union[str, sa.ColumnElement[bool]]]] = [
    (
        "target_reps_type_integer_or_null",
        "typeof(target_reps) = 'integer' or typeof(target_reps) = 'null'",
    ),
    (
        "target_time_type_integer_or_null",
        "typeof(target_time) = 'integer' or typeof(target_time) = 'null'",
    ),
    (
        "target_weight_type_real_or_null",
        "typeof(target_weight) = 'real' or typeof(target_weight) = 'null'",
    ),
    ("target_rpe_type_real_or_null", "typeof(target_rpe) = 'real' or typeof(target_rpe) = 'null'"),
    ("target_reps_gt_0", sa.column("target_reps") > 0),
    ("target_time_gt_0", sa.column("target_time") > 0),
    ("target_weight_gt_0", sa.column("target_weight") > 0),
    ("target_rpe_ge_0", sa.column("target_rpe") >= 0),
    ("target_rpe_le_10", sa.column("target_rpe") <= 10),
]


def upgrade() -> None:
    op.create_table(
        "workout_element",
        sa.Column("workout_id", sa.Integer(), nullable=False),
        sa.Column("position", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(), nullable=False),
        sa.Column("automatic", sa.Boolean(), nullable=False),
        sa.CheckConstraint(
            "typeof(position) = 'integer'", name=op.f("ck_workout_element_position_type_integer")
        ),
        sa.CheckConstraint("typeof(type) = 'text'", name=op.f("ck_workout_element_type_type_text")),
        sa.CheckConstraint(
            "typeof(automatic) = 'integer'", name=op.f("ck_workout_element_automatic_type_integer")
        ),
        sa.CheckConstraint("automatic <= 1", name=op.f("ck_workout_element_rpe_le_1")),
        sa.CheckConstraint("automatic >= 0", name=op.f("ck_workout_element_rpe_ge_0")),
        sa.CheckConstraint("position > 0", name=op.f("ck_workout_element_position_gt_0")),
        sa.ForeignKeyConstraint(
            ["workout_id"],
            ["workout.id"],
            name=op.f("fk_workout_element_workout_id_workout"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("workout_id", "position", name=op.f("pk_workout_element")),
    )

    connection = op.get_bind()

    for ws in connection.execute(workout_set.select()):
        connection.execute(
            workout_element.insert().values(
                workout_id=ws.workout_id,
                position=ws.position,
                type="set",
                automatic=False,
            )
        )

    op.create_table(
        "workout_rest",
        sa.Column("workout_id", sa.Integer(), nullable=False),
        sa.Column("position", sa.Integer(), nullable=False),
        sa.Column("target_time", sa.Integer(), nullable=True),
        sa.CheckConstraint(
            "typeof(position) = 'integer'", name=op.f("ck_workout_rest_position_type_integer")
        ),
        sa.CheckConstraint(
            "typeof(target_time) = 'integer' or typeof(target_time) = 'null'",
            name=op.f("ck_workout_rest_target_time_type_integer_or_null"),
        ),
        sa.CheckConstraint("position > 0", name=op.f("ck_workout_rest_position_gt_0")),
        sa.CheckConstraint("target_time > 0", name=op.f("ck_workout_rest_target_time_gt_0")),
        sa.ForeignKeyConstraint(
            ["workout_id", "position"],
            ["workout_element.workout_id", "workout_element.position"],
            name=op.f("fk_workout_rest_workout_id_workout_element"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("workout_id", "position", name=op.f("pk_workout_rest")),
    )

    with op.batch_alter_table("workout_set", schema=None) as batch_op:
        batch_op.add_column(sa.Column("target_reps", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("target_time", sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column("target_weight", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("target_rpe", sa.Float(), nullable=True))
        batch_op.drop_constraint("fk_workout_set_workout_id_workout", type_="foreignkey")
        batch_op.create_foreign_key(
            batch_op.f("fk_workout_set_workout_id_workout_element"),
            "workout_element",
            ["workout_id", "position"],
            ["workout_id", "position"],
            ondelete="CASCADE",
        )
        for constraint_name, condition in check_constraints:
            batch_op.create_check_constraint(constraint_name, condition)


def downgrade() -> None:
    with op.batch_alter_table("workout_set", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_workout_set_workout_id_workout_element"), type_="foreignkey"
        )
        batch_op.create_foreign_key(
            "fk_workout_set_workout_id_workout",
            "workout",
            ["workout_id"],
            ["id"],
            ondelete="CASCADE",
        )
        for constraint_name, _ in check_constraints:
            batch_op.drop_constraint(constraint_name, type_="check")
        batch_op.drop_column("target_rpe")
        batch_op.drop_column("target_weight")
        batch_op.drop_column("target_time")
        batch_op.drop_column("target_reps")

    op.drop_table("workout_rest")
    op.drop_table("workout_element")



"""
Add routine sections.

Revision ID: b9f4e42c7135
Revises: 4b6051594962
Create Date: 2022-11-13

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy import Boolean, Integer, String
from sqlalchemy.sql import column, table

revision = "b9f4e42c7135"
down_revision = "4b6051594962"
branch_labels = None
depends_on = None


routine_exercise = table(
    "routine_exercise",
    column("routine_id", Integer),
    column("position", Integer),
    column("exercise_id", Integer),
    column("sets", Integer),
)
routine_part = table(
    "routine_part",
    column("id", Integer),
    column("type", String),
    column("routine_section_id", Integer),
    column("position", Integer),
)
routine_section = table(
    "routine_section",
    column("id", Integer),
    column("routine_id", Integer),
    column("rounds", Integer),
)
routine_activity = table(
    "routine_activity",
    column("id", Integer),
    column("exercise_id", Integer),
    column("duration", Integer),
    column("tempo", Integer),
    column("automatic", Boolean),
)


def upgrade() -> None:
    op.create_table(
        "routine_part",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("type", sa.String(), nullable=False),
        sa.Column("routine_section_id", sa.Integer(), nullable=True),
        sa.Column("position", sa.Integer(), nullable=False),
        sa.CheckConstraint(
            "typeof(position) = 'integer'", name=op.f("ck_routine_part_position_type_integer")
        ),
        sa.CheckConstraint("typeof(type) = 'text'", name=op.f("ck_routine_part_type_type_text")),
        sa.CheckConstraint("position > 0", name=op.f("ck_routine_part_position_gt_0")),
        sa.ForeignKeyConstraint(
            ["routine_section_id"],
            ["routine_section.id"],
            name=op.f("fk_routine_part_routine_section_id_routine_section"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_routine_part")),
    )
    op.create_table(
        "routine_section",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("routine_id", sa.Integer(), nullable=True),
        sa.Column("rounds", sa.Integer(), nullable=False),
        sa.CheckConstraint(
            "typeof(rounds) = 'integer'", name=op.f("ck_routine_section_rounds_type_integer")
        ),
        sa.CheckConstraint("rounds > 0", name=op.f("ck_routine_section_rounds_gt_0")),
        sa.ForeignKeyConstraint(
            ["id"], ["routine_part.id"], name=op.f("fk_routine_section_id_routine_part")
        ),
        sa.ForeignKeyConstraint(
            ["routine_id"],
            ["routine.id"],
            name=op.f("fk_routine_section_routine_id_routine"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_routine_section")),
    )
    op.create_table(
        "routine_activity",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("exercise_id", sa.Integer(), nullable=True),
        sa.Column("duration", sa.Integer(), nullable=False),
        sa.Column("tempo", sa.Integer(), nullable=False),
        sa.Column("automatic", sa.Boolean(), nullable=False),
        sa.CheckConstraint(
            "typeof(duration) = 'integer'", name=op.f("ck_routine_activity_duration_type_integer")
        ),
        sa.CheckConstraint("duration >= 0", name=op.f("ck_routine_activity_duration_ge_0")),
        sa.CheckConstraint(
            "typeof(tempo) = 'integer'", name=op.f("ck_routine_activity_tempo_type_integer")
        ),
        sa.CheckConstraint("tempo >= 0", name=op.f("ck_routine_activity_tempo_ge_0")),
        sa.ForeignKeyConstraint(
            ["id"], ["routine_part.id"], name=op.f("fk_routine_activity_id_routine_part")
        ),
        sa.ForeignKeyConstraint(
            ["exercise_id"],
            ["exercise.id"],
            name=op.f("fk_routine_activity_exercise_id_exercise"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_routine_activity")),
    )

    connection = op.get_bind()

    part_id = 1

    for exercise in connection.execute(routine_exercise.select()):
        connection.execute(
            routine_part.insert().values(
                id=part_id, type="routine_section", position=exercise.position
            )
        )
        connection.execute(
            routine_section.insert().values(
                id=part_id, routine_id=exercise.routine_id, rounds=exercise.sets
            )
        )

        connection.execute(
            routine_part.insert().values(
                id=part_id + 1, type="routine_activity", routine_section_id=part_id, position=1
            )
        )
        connection.execute(
            routine_activity.insert().values(
                id=part_id + 1,
                exercise_id=exercise.exercise_id,
                duration=0,
                tempo=0,
                automatic=False,
            )
        )
        connection.execute(
            routine_part.insert().values(
                id=part_id + 2, type="routine_activity", routine_section_id=part_id, position=1
            )
        )
        connection.execute(
            routine_activity.insert().values(
                id=part_id + 2, exercise_id=None, duration=60, tempo=0, automatic=True
            )
        )

        part_id += 3

    op.drop_table("routine_exercise")


def downgrade() -> None:
    op.create_table(
        "routine_exercise",
        sa.Column("routine_id", sa.INTEGER(), nullable=False),
        sa.Column("position", sa.INTEGER(), nullable=False),
        sa.Column("exercise_id", sa.INTEGER(), nullable=False),
        sa.Column("sets", sa.INTEGER(), nullable=False),
        sa.CheckConstraint("typeof(position) = 'integer'", name="position_type_integer"),
        sa.CheckConstraint("typeof(sets) = 'integer'", name="sets_type_integer"),
        sa.CheckConstraint("position > 0", name="position_gt_0"),
        sa.CheckConstraint("sets > 0", name="sets_gt_0"),
        sa.ForeignKeyConstraint(
            ["routine_id"],
            ["routine.id"],
            name="fk_routine_exercise_routine_id_routine",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["exercise_id"],
            ["exercise.id"],
            name="fk_routine_exercise_exercise_id_exercise",
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("routine_id", "position", name="pk_routine_exercise"),
    )

    connection = op.get_bind()

    for section in connection.execute(
        routine_part.join(routine_section, routine_part.c.id == routine_section.c.id).select()
    ):
        for activity in connection.execute(
            routine_part.join(routine_activity, routine_part.c.id == routine_activity.c.id)
            .select()
            .where(routine_part.c.routine_section_id == section.id)
            .where(routine_activity.c.exercise_id != None)  # noqa: E711
        ):
            if section.routine_id is None:
                continue
            connection.execute(
                routine_exercise.insert().values(
                    routine_id=section.routine_id,
                    position=section.position,
                    exercise_id=activity.exercise_id,
                    sets=section.rounds,
                )
            )

    op.drop_table("routine_activity")
    op.drop_table("routine_section")
    op.drop_table("routine_part")



"""
Add routine_id to workout.

Revision ID: 4cacd61cb0c5
Revises:
Create Date: 2021-07-03 21:19:40.634728

"""

import sqlalchemy as sa
from alembic import op

revision = "4cacd61cb0c5"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("workout", schema=None) as batch_op:
        batch_op.add_column(sa.Column("routine_id", sa.Integer(), nullable=True))
        batch_op.create_foreign_key(
            batch_op.f("fk_workout_routine_id_routine"), "routine", ["routine_id"], ["id"]
        )


def downgrade() -> None:
    with op.batch_alter_table("workout", schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f("fk_workout_routine_id_routine"), type_="foreignkey")
        batch_op.drop_column("routine_id")



from __future__ import annotations

import os
from pathlib import Path

from flask import current_app


def check_app_config() -> None:
    for key in ["DATABASE", "SECRET_KEY"]:
        if key not in current_app.config:
            raise RuntimeError(f"'{key}' is not set in app config")


def check_config_file(environ: dict[str, str]) -> None:
    if "VALENS_CONFIG" not in environ:
        raise RuntimeError("environment variable 'VALENS_CONFIG' is not set")

    config_file = Path(environ["VALENS_CONFIG"])

    if not config_file.exists():
        raise RuntimeError(f"config file '{config_file}' not found")

    check_app_config()


def create_config_file(config_directory: Path, database_file: Path) -> Path:
    config = config_directory / "config.py"
    config.write_text(
        f"DATABASE = 'sqlite:///{database_file}'\nSECRET_KEY = {os.urandom(24)!r}\n",
        encoding="utf-8",
    )
    return config



from __future__ import annotations

import datetime
import enum
from typing import Optional

from sqlalchemy import (
    CheckConstraint,
    Constraint,
    Date,
    Enum,
    Float,
    ForeignKey,
    ForeignKeyConstraint,
    Integer,
    MetaData,
    String,
    UniqueConstraint,
    column,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

# Alembic's autogenerate does not support CHECK constraints
# (https://github.com/sqlalchemy/alembic/issues/508). CHECK constraints get lost when running
# "batch" migrations for SQLite and must be stated explicitly in the migration script
# (https://alembic.sqlalchemy.org/en/latest/batch.html#including-check-constraints).
# CHECK constraints are only specified at the table level to enable the schema to be kept identical
# when applying migrations despite the aforementioned limitations.


class Base(DeclarativeBase):
    metadata = MetaData(
        naming_convention={
            "ix": "ix_%(column_0_label)s",
            "uq": "uq_%(table_name)s_%(column_0_name)s",
            "ck": "ck_%(table_name)s_%(constraint_name)s",
            "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
            "pk": "pk_%(table_name)s",
        }
    )


class Sex(enum.IntEnum):
    FEMALE = 0
    MALE = 1


class User(Base):
    __tablename__ = "user"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    sex: Mapped[Sex] = mapped_column(Enum(Sex), nullable=False)

    body_weight: Mapped[list[BodyWeight]] = relationship(
        "BodyWeight", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )
    body_fat: Mapped[list[BodyFat]] = relationship(
        "BodyFat", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )
    period: Mapped[list[Period]] = relationship(
        "Period", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )
    exercises: Mapped[list[Exercise]] = relationship(
        "Exercise", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )
    routines: Mapped[list[Routine]] = relationship(
        "Routine", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )
    workouts: Mapped[list[Workout]] = relationship(
        "Workout", backref="user", cascade="all, delete-orphan", passive_deletes=True
    )


class BodyWeight(Base):
    __tablename__ = "body_weight"
    __table_args__ = (
        CheckConstraint("typeof(weight) = 'real'", name="weight_type_real"),
        CheckConstraint(column("weight") > 0, name="weight_gt_0"),
    )

    user_id: Mapped[int] = mapped_column(
        ForeignKey("user.id", ondelete="CASCADE"), primary_key=True
    )
    date: Mapped[datetime.date] = mapped_column(Date, primary_key=True)
    weight: Mapped[float] = mapped_column(Float, nullable=False)


class BodyFat(Base):
    __tablename__ = "body_fat"
    __table_args__ = (
        CheckConstraint(
            "typeof(chest) = 'integer' or typeof(chest) = 'null'",
            name="chest_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(abdominal) = 'integer' or typeof(abdominal) = 'null'",
            name="abdominal_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(tigh) = 'integer' or typeof(tigh) = 'null'",
            name="tigh_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(tricep) = 'integer' or typeof(tricep) = 'null'",
            name="tricep_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(subscapular) = 'integer' or typeof(subscapular) = 'null'",
            name="subscapular_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(suprailiac) = 'integer' or typeof(suprailiac) = 'null'",
            name="suprailiac_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(midaxillary) = 'integer' or typeof(midaxillary) = 'null'",
            name="midaxillary_type_integer_or_null",
        ),
        CheckConstraint(column("chest") > 0, name="chest_gt_0"),
        CheckConstraint(column("abdominal") > 0, name="abdominal_gt_0"),
        CheckConstraint(column("tigh") > 0, name="tigh_gt_0"),
        CheckConstraint(column("tricep") > 0, name="tricep_gt_0"),
        CheckConstraint(column("subscapular") > 0, name="subscapular_gt_0"),
        CheckConstraint(column("suprailiac") > 0, name="suprailiac_gt_0"),
        CheckConstraint(column("midaxillary") > 0, name="midaxillary_gt_0"),
    )

    user_id: Mapped[int] = mapped_column(
        ForeignKey("user.id", ondelete="CASCADE"), primary_key=True
    )
    date: Mapped[datetime.date] = mapped_column(Date, primary_key=True)
    chest: Mapped[Optional[int]] = mapped_column(Integer)
    abdominal: Mapped[Optional[int]] = mapped_column(Integer)
    tigh: Mapped[Optional[int]] = mapped_column(Integer)
    tricep: Mapped[Optional[int]] = mapped_column(Integer)
    subscapular: Mapped[Optional[int]] = mapped_column(Integer)
    suprailiac: Mapped[Optional[int]] = mapped_column(Integer)
    midaxillary: Mapped[Optional[int]] = mapped_column(Integer)


class Period(Base):
    __tablename__ = "period"
    __table_args__ = (
        CheckConstraint("typeof(intensity) = 'integer'", name="intensity_type_integer"),
        CheckConstraint(column("intensity") >= 1, name="intensity_ge_1"),
        CheckConstraint(column("intensity") <= 4, name="intensity_le_4"),
    )

    user_id: Mapped[int] = mapped_column(
        ForeignKey("user.id", ondelete="CASCADE"), primary_key=True
    )
    date: Mapped[datetime.date] = mapped_column(Date, primary_key=True)
    intensity: Mapped[int] = mapped_column(Integer, nullable=False)


class Exercise(Base):
    __tablename__ = "exercise"
    __table_args__ = (UniqueConstraint("user_id", "name"),)

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id", ondelete="CASCADE"), nullable=False)
    name: Mapped[str] = mapped_column(String, nullable=False)

    sets: Mapped[list[WorkoutSet]] = relationship(
        "WorkoutSet", back_populates="exercise", cascade="all, delete-orphan"
    )
    routine_activities: Mapped[list[RoutineActivity]] = relationship(
        "RoutineActivity", back_populates="exercise", cascade="all, delete-orphan"
    )


class Routine(Base):
    __tablename__ = "routine"
    __table_args__ = (UniqueConstraint("user_id", "name"),)

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id", ondelete="CASCADE"), nullable=False)
    name: Mapped[str] = mapped_column(String, nullable=False)
    notes: Mapped[Optional[str]] = mapped_column(String)

    sections: Mapped[list[RoutineSection]] = relationship(
        "RoutineSection", back_populates="routine", cascade="all, delete-orphan"
    )
    workouts: Mapped[list[Workout]] = relationship("Workout", back_populates="routine")


class RoutinePart(Base):
    __tablename__ = "routine_part"
    __table_args__: tuple[CheckConstraint, ...] = (
        CheckConstraint("typeof(position) = 'integer'", name="position_type_integer"),
        CheckConstraint("typeof(type) = 'text'", name="type_type_text"),
        CheckConstraint(column("position") > 0, name="position_gt_0"),
    )

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    type: Mapped[str] = mapped_column(String, nullable=False)
    routine_section_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("routine_section.id", ondelete="CASCADE")
    )
    position: Mapped[int] = mapped_column(Integer, nullable=False)

    section: Mapped[RoutineSection] = relationship(
        "RoutineSection",
        back_populates="parts",
        foreign_keys=[routine_section_id],
    )

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "routine_part",
        "polymorphic_on": type,
    }


class RoutineSection(RoutinePart):
    __tablename__ = "routine_section"
    __table_args__ = (
        CheckConstraint("typeof(rounds) = 'integer'", name="rounds_type_integer"),
        CheckConstraint(column("rounds") > 0, name="rounds_gt_0"),
    )

    id: Mapped[int] = mapped_column(Integer, ForeignKey("routine_part.id"), primary_key=True)
    routine_id: Mapped[Optional[int]] = mapped_column(ForeignKey("routine.id", ondelete="CASCADE"))
    rounds: Mapped[int] = mapped_column(Integer, nullable=False)

    parts: Mapped[list[RoutinePart]] = relationship(
        "RoutinePart",
        back_populates="section",
        foreign_keys=RoutinePart.routine_section_id,
    )
    routine: Mapped[Routine] = relationship("Routine", back_populates="sections")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "routine_section",
        "polymorphic_load": "selectin",
        "inherit_condition": id == RoutinePart.id,
    }


class RoutineActivity(RoutinePart):
    __tablename__ = "routine_activity"
    __table_args__ = (
        CheckConstraint("typeof(reps) = 'integer'", name="reps_type_integer"),
        CheckConstraint(column("reps") >= 0, name="reps_ge_0"),
        CheckConstraint("typeof(time) = 'integer'", name="time_type_integer"),
        CheckConstraint(column("time") >= 0, name="time_ge_0"),
        CheckConstraint("typeof(weight) = 'real'", name="weight_type_real"),
        CheckConstraint(column("weight") >= 0, name="weight_ge_0"),
        CheckConstraint("typeof(rpe) = 'real'", name="rpe_type_real"),
        CheckConstraint(column("rpe") >= 0, name="rpe_ge_0"),
        CheckConstraint(column("rpe") <= 10, name="rpe_le_10"),
    )

    id: Mapped[int] = mapped_column(Integer, ForeignKey("routine_part.id"), primary_key=True)
    exercise_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("exercise.id", ondelete="CASCADE")
    )
    reps: Mapped[int]
    time: Mapped[int]
    weight: Mapped[float]
    rpe: Mapped[float]
    automatic: Mapped[bool]

    exercise: Mapped[Exercise] = relationship("Exercise", back_populates="routine_activities")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "routine_activity",
        "polymorphic_load": "selectin",
        "inherit_condition": id == RoutinePart.id,
    }


class Workout(Base):
    __tablename__ = "workout"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(ForeignKey("user.id", ondelete="CASCADE"), nullable=False)
    routine_id: Mapped[Optional[int]] = mapped_column(ForeignKey("routine.id", ondelete="CASCADE"))
    date: Mapped[datetime.date] = mapped_column(Date, nullable=False)
    notes: Mapped[Optional[str]] = mapped_column(String)

    routine: Mapped[Routine] = relationship("Routine", back_populates="workouts")
    elements: Mapped[list[WorkoutElement]] = relationship(
        "WorkoutElement", back_populates="workout", cascade="all, delete-orphan"
    )


class WorkoutElement(Base):
    __tablename__ = "workout_element"
    __table_args__: tuple[Constraint, ...] = (
        CheckConstraint(
            "typeof(position) = 'integer'",
            name="position_type_integer",
        ),
        CheckConstraint("typeof(type) = 'text'", name="type_type_text"),
        CheckConstraint(
            "typeof(automatic) = 'integer'",
            name="automatic_type_integer",
        ),
        CheckConstraint(column("position") > 0, name="position_gt_0"),
        CheckConstraint(column("automatic") >= 0, name="rpe_ge_0"),
        CheckConstraint(column("automatic") <= 1, name="rpe_le_1"),
    )

    workout_id: Mapped[int] = mapped_column(
        ForeignKey("workout.id", ondelete="CASCADE"), primary_key=True
    )
    position: Mapped[int] = mapped_column(primary_key=True)
    type: Mapped[str]
    automatic: Mapped[bool] = mapped_column(default=False)

    workout: Mapped[Workout] = relationship("Workout", back_populates="elements")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "activity",
        "polymorphic_on": "type",
    }


class WorkoutSet(WorkoutElement):
    __tablename__ = "workout_set"
    __table_args__ = (
        CheckConstraint(
            "typeof(position) = 'integer'",
            name="position_type_integer",
        ),
        CheckConstraint(
            "typeof(reps) = 'integer' or typeof(reps) = 'null'",
            name="reps_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(time) = 'integer' or typeof(time) = 'null'",
            name="time_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(weight) = 'real' or typeof(weight) = 'null'",
            name="weight_type_real_or_null",
        ),
        CheckConstraint(
            "typeof(rpe) = 'real' or typeof(rpe) = 'null'",
            name="rpe_type_real_or_null",
        ),
        CheckConstraint(
            "typeof(target_reps) = 'integer' or typeof(target_reps) = 'null'",
            name="target_reps_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(target_time) = 'integer' or typeof(target_time) = 'null'",
            name="target_time_type_integer_or_null",
        ),
        CheckConstraint(
            "typeof(target_weight) = 'real' or typeof(target_weight) = 'null'",
            name="target_weight_type_real_or_null",
        ),
        CheckConstraint(
            "typeof(target_rpe) = 'real' or typeof(target_rpe) = 'null'",
            name="target_rpe_type_real_or_null",
        ),
        CheckConstraint(column("position") > 0, name="position_gt_0"),
        CheckConstraint(column("reps") > 0, name="reps_gt_0"),
        CheckConstraint(column("time") > 0, name="time_gt_0"),
        CheckConstraint(column("weight") > 0, name="weight_gt_0"),
        CheckConstraint(column("rpe") >= 0, name="rpe_ge_0"),
        CheckConstraint(column("rpe") <= 10, name="rpe_le_10"),
        CheckConstraint(column("target_reps") > 0, name="target_reps_gt_0"),
        CheckConstraint(column("target_time") > 0, name="target_time_gt_0"),
        CheckConstraint(column("target_weight") > 0, name="target_weight_gt_0"),
        CheckConstraint(column("target_rpe") >= 0, name="target_rpe_ge_0"),
        CheckConstraint(column("target_rpe") <= 10, name="target_rpe_le_10"),
        ForeignKeyConstraint(
            ["workout_id", "position"],
            [WorkoutElement.workout_id, WorkoutElement.position],
            ondelete="CASCADE",
        ),
    )

    workout_id: Mapped[int] = mapped_column(primary_key=True)
    position: Mapped[int] = mapped_column(primary_key=True)
    exercise_id: Mapped[int] = mapped_column(
        ForeignKey("exercise.id", ondelete="CASCADE"), nullable=False
    )
    reps: Mapped[Optional[int]]
    time: Mapped[Optional[int]]
    weight: Mapped[Optional[float]]
    rpe: Mapped[Optional[float]]
    target_reps: Mapped[Optional[int]]
    target_time: Mapped[Optional[int]]
    target_weight: Mapped[Optional[float]]
    target_rpe: Mapped[Optional[float]]

    exercise: Mapped[Exercise] = relationship("Exercise", back_populates="sets")

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "set",
        "polymorphic_load": "selectin",
    }


class WorkoutRest(WorkoutElement):
    __tablename__ = "workout_rest"
    __table_args__ = (
        CheckConstraint(
            "typeof(position) = 'integer'",
            name="position_type_integer",
        ),
        CheckConstraint(
            "typeof(target_time) = 'integer' or typeof(target_time) = 'null'",
            name="target_time_type_integer_or_null",
        ),
        CheckConstraint(column("position") > 0, name="position_gt_0"),
        CheckConstraint(column("target_time") > 0, name="target_time_gt_0"),
        ForeignKeyConstraint(
            ["workout_id", "position"],
            [WorkoutElement.workout_id, WorkoutElement.position],
            ondelete="CASCADE",
        ),
    )

    workout_id: Mapped[int] = mapped_column(primary_key=True)
    position: Mapped[int] = mapped_column(primary_key=True)
    target_time: Mapped[Optional[int]]

    __mapper_args__ = {  # noqa: RUF012
        "polymorphic_identity": "rest",
        "polymorphic_load": "selectin",
    }



from __future__ import annotations

from datetime import date
from functools import singledispatch, wraps
from http import HTTPStatus
from itertools import chain
from typing import Any, Callable, Optional

from flask import Blueprint, jsonify, request, session
from flask.typing import ResponseReturnValue
from sqlalchemy import column, select
from sqlalchemy.exc import IntegrityError, NoResultFound
from sqlalchemy.orm import selectinload

from valens import database as db, version
from valens.models import (
    BodyFat,
    BodyWeight,
    Exercise,
    Period,
    Routine,
    RoutineActivity,
    RoutinePart,
    RoutineSection,
    Sex,
    User,
    Workout,
    WorkoutElement,
    WorkoutRest,
    WorkoutSet,
)

bp = Blueprint("api", __name__, url_prefix="/api")


class DeserializationError(Exception):
    pass


@singledispatch
def to_dict(
    model: object, exclude: Optional[list[str]] = None, include: Optional[list[str]] = None
) -> dict[str, object]:
    return model_to_dict(model, exclude, include)


@to_dict.register
def _(model: Routine) -> dict[str, object]:
    return {
        **model_to_dict(model),
        "sections": [to_dict(s) for s in sorted(model.sections, key=lambda x: x.position)],
    }


@to_dict.register
def _(model: RoutineSection) -> dict[str, object]:
    return {
        **model_to_dict(model, exclude=["id", "routine_id"]),
        "parts": [to_dict(p) for p in sorted(model.parts, key=lambda x: x.position)],
    }


@to_dict.register
def _(model: RoutineActivity) -> dict[str, object]:
    return {
        **model_to_dict(model, exclude=["id"]),
    }


@to_dict.register
def _(model: Workout) -> dict[str, object]:
    return {
        **model_to_dict(model),
        "elements": [to_dict(e) for e in model.elements],
    }


@to_dict.register
def _(model: WorkoutElement) -> dict[str, object]:
    return {
        **model_to_dict(model, exclude=["workout_id", "position"], include=["automatic"]),
    }


def model_to_dict(
    model: object, exclude: Optional[list[str]] = None, include: Optional[list[str]] = None
) -> dict[str, object]:
    assert hasattr(model, "__table__")
    exclude = ["user_id"] if exclude is None else exclude
    include = [] if include is None else include
    return {
        name: attr.isoformat() if isinstance(attr, date) else attr
        for col in chain(model.__table__.columns, (column(i) for i in include))
        if col.name not in exclude
        for name, attr in [(col.name, getattr(model, col.name))]
    }


def to_routine_parts(json: list[dict[str, Any]]) -> list[RoutinePart]:  # type: ignore[misc]
    return [
        (
            to_routine_section(part, position)
            if "rounds" in part
            else to_routine_activity(part, position)
        )
        for position, part in enumerate(json, start=1)
    ]


def to_routine_sections(json: list[dict[str, Any]]) -> list[RoutineSection]:  # type: ignore[misc]
    return [to_routine_section(section, position) for position, section in enumerate(json, start=1)]


def to_routine_section(json: dict[str, Any], position: int) -> RoutineSection:  # type: ignore[misc]
    return RoutineSection(
        position=position,
        rounds=json["rounds"],
        parts=to_routine_parts(json["parts"]),
    )


def to_routine_activity(  # type: ignore[misc]
    json: dict[str, Any], position: int
) -> RoutineActivity:
    return RoutineActivity(
        position=position,
        exercise_id=json["exercise_id"],
        reps=json["reps"],
        time=json["time"],
        weight=json["weight"],
        rpe=json["rpe"],
        automatic=json["automatic"],
    )


def to_workout_elements(json: list[dict[str, Any]]) -> list[WorkoutElement]:  # type: ignore[misc]
    return [
        (
            WorkoutSet(
                position=position,
                exercise_id=element["exercise_id"],
                reps=element["reps"],
                time=element["time"],
                weight=element["weight"],
                rpe=element["rpe"],
                target_reps=element["target_reps"],
                target_time=element["target_time"],
                target_weight=element["target_weight"],
                target_rpe=element["target_rpe"],
                automatic=element["automatic"],
            )
            if "exercise_id" in element
            else WorkoutRest(
                position=position,
                target_time=element["target_time"],
                automatic=element["automatic"],
            )
        )
        for position, element in enumerate(json, start=1)
    ]


def json_expected(function: Callable) -> Callable:  # type: ignore[type-arg]
    @wraps(function)
    def decorated_function(*args: object, **kwargs: object) -> ResponseReturnValue:
        if not request.is_json:
            return "", HTTPStatus.UNSUPPORTED_MEDIA_TYPE
        return function(*args, **kwargs)

    return decorated_function


def session_required(function: Callable) -> Callable:  # type: ignore[type-arg]
    @wraps(function)
    def decorated_function(*args: object, **kwargs: object) -> ResponseReturnValue:
        if "username" not in session or "user_id" not in session or "sex" not in session:
            return "", HTTPStatus.UNAUTHORIZED
        return function(*args, **kwargs)

    return decorated_function


@bp.route("/version")
def read_version() -> ResponseReturnValue:
    return jsonify(version.get())


@bp.route("/session")
def read_session() -> ResponseReturnValue:
    if "username" not in session or "user_id" not in session or "sex" not in session:
        return "", HTTPStatus.NOT_FOUND

    return jsonify({"id": session["user_id"], "name": session["username"], "sex": session["sex"]})


@bp.route("/session", methods=["POST"])
@json_expected
def create_session() -> ResponseReturnValue:
    try:
        assert isinstance(request.json, dict)
        user_id = request.json["id"]
    except KeyError as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        user = db.session.execute(select(User).where(User.id == user_id)).scalars().one()
    except NoResultFound:
        return "", HTTPStatus.NOT_FOUND

    session["user_id"] = user.id
    session["username"] = user.name
    session["sex"] = user.sex
    session.permanent = True

    return jsonify(to_dict(user))


@bp.route("/session", methods=["DELETE"])
def delete_session() -> ResponseReturnValue:
    session.clear()
    return "", HTTPStatus.NO_CONTENT


@bp.route("/users")
def read_users() -> ResponseReturnValue:
    users = db.session.execute(select(User)).scalars().all()
    return jsonify([to_dict(u) for u in users])


@bp.route("/users/<int:user_id>")
@session_required
def read_user(user_id: int) -> ResponseReturnValue:
    try:
        user = db.session.execute(select(User).where(User.id == user_id)).scalars().one()
    except NoResultFound:
        return "", HTTPStatus.NOT_FOUND

    return jsonify(to_dict(user))


@bp.route("/users", methods=["POST"])
@json_expected
def create_user() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        user = User(name=data["name"].strip(), sex=Sex(data["sex"]))
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(user)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(user)),
        HTTPStatus.CREATED,
        {"Location": f"/users/{user.id}"},
    )


@bp.route("/users/<int:user_id>", methods=["PUT"])
@json_expected
def replace_user(user_id: int) -> ResponseReturnValue:
    try:
        user = db.session.execute(select(User).where(User.id == user_id)).scalars().one()
    except NoResultFound:
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        user.name = data["name"].strip()
        user.sex = Sex(data["sex"])
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return jsonify(to_dict(user)), HTTPStatus.OK


@bp.route("/users/<int:user_id>", methods=["DELETE"])
def delete_user(user_id: int) -> ResponseReturnValue:
    try:
        user = db.session.execute(select(User).where(User.id == user_id)).scalars().one()
    except NoResultFound:
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(user)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/body_weight")
@session_required
def read_body_weight() -> ResponseReturnValue:
    body_weight = (
        db.session.execute(select(BodyWeight).where(BodyWeight.user_id == session["user_id"]))
        .scalars()
        .all()
    )
    return jsonify([to_dict(bw) for bw in body_weight])


@bp.route("/body_weight", methods=["POST"])
@session_required
@json_expected
def create_body_weight() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        body_weight = BodyWeight(
            user_id=session["user_id"],
            date=date.fromisoformat(data["date"]),
            weight=float(data["weight"]),
        )
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(body_weight)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(body_weight)),
        HTTPStatus.CREATED,
        {"Location": f"/body_weight/{body_weight.date}"},
    )


@bp.route("/body_weight/<date_>", methods=["PUT"])
@session_required
@json_expected
def replace_body_weight(date_: str) -> ResponseReturnValue:
    try:
        body_weight = (
            db.session.execute(
                select(BodyWeight)
                .where(BodyWeight.user_id == session["user_id"])
                .where(BodyWeight.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        body_weight.weight = float(data["weight"])
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(body_weight)),
        HTTPStatus.OK,
    )


@bp.route("/body_weight/<date_>", methods=["DELETE"])
@session_required
def delete_body_weight(date_: str) -> ResponseReturnValue:
    try:
        body_weight = (
            db.session.execute(
                select(BodyWeight)
                .where(BodyWeight.user_id == session["user_id"])
                .where(BodyWeight.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(body_weight)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/body_fat")
@session_required
def read_body_fat() -> ResponseReturnValue:
    body_fat = (
        db.session.execute(select(BodyFat).where(BodyFat.user_id == session["user_id"]))
        .scalars()
        .all()
    )
    return jsonify([to_dict(bf) for bf in body_fat])


@bp.route("/body_fat", methods=["POST"])
@session_required
@json_expected
def create_body_fat() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        body_fat = BodyFat(
            user_id=int(session["user_id"]),
            date=date.fromisoformat(data["date"]),
            **{
                part: int(data[part]) if data[part] is not None else None
                for part in [
                    "chest",
                    "abdominal",
                    "tigh",
                    "tricep",
                    "subscapular",
                    "suprailiac",
                    "midaxillary",
                ]
            },
        )
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(body_fat)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(body_fat)),
        HTTPStatus.CREATED,
        {"Location": f"/body_fat/{body_fat.date}"},
    )


@bp.route("/body_fat/<date_>", methods=["PUT"])
@session_required
@json_expected
def replace_body_fat(date_: str) -> ResponseReturnValue:
    try:
        body_fat = (
            db.session.execute(
                select(BodyFat)
                .where(BodyFat.user_id == session["user_id"])
                .where(BodyFat.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        for attr in [
            "chest",
            "abdominal",
            "tigh",
            "tricep",
            "subscapular",
            "suprailiac",
            "midaxillary",
        ]:
            setattr(body_fat, attr, int(data[attr]) if data[attr] is not None else None)
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(body_fat)),
        HTTPStatus.OK,
    )


@bp.route("/body_fat/<date_>", methods=["DELETE"])
@session_required
def delete_body_fat(date_: str) -> ResponseReturnValue:
    try:
        body_fat = (
            db.session.execute(
                select(BodyFat)
                .where(BodyFat.user_id == session["user_id"])
                .where(BodyFat.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(body_fat)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/period")
@session_required
def read_period() -> ResponseReturnValue:
    period = (
        db.session.execute(select(Period).where(Period.user_id == session["user_id"]))
        .scalars()
        .all()
    )
    return jsonify([to_dict(p) for p in period])


@bp.route("/period", methods=["POST"])
@session_required
@json_expected
def create_period() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        period = Period(
            user_id=session["user_id"],
            date=date.fromisoformat(data["date"]),
            intensity=int(data["intensity"]),
        )
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(period)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(period)),
        HTTPStatus.CREATED,
        {"Location": f"/period/{period.date}"},
    )


@bp.route("/period/<date_>", methods=["PUT"])
@session_required
@json_expected
def replace_period(date_: str) -> ResponseReturnValue:
    try:
        period = (
            db.session.execute(
                select(Period)
                .where(Period.user_id == session["user_id"])
                .where(Period.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        period.intensity = int(data["intensity"])
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(period)),
        HTTPStatus.OK,
    )


@bp.route("/period/<date_>", methods=["DELETE"])
@session_required
def delete_period(date_: str) -> ResponseReturnValue:
    try:
        period = (
            db.session.execute(
                select(Period)
                .where(Period.user_id == session["user_id"])
                .where(Period.date == date.fromisoformat(date_))
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(period)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/exercises")
@session_required
def read_exercises() -> ResponseReturnValue:
    exercises = (
        db.session.execute(select(Exercise).where(Exercise.user_id == session["user_id"]))
        .scalars()
        .all()
    )
    return jsonify([to_dict(e) for e in exercises])


@bp.route("/exercises", methods=["POST"])
@session_required
@json_expected
def create_exercise() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        exercise = Exercise(
            user_id=session["user_id"],
            name=data["name"],
        )
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(exercise)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(exercise)),
        HTTPStatus.CREATED,
        {"Location": f"/exercises/{exercise.id}"},
    )


@bp.route("/exercises/<int:id_>", methods=["PUT"])
@session_required
@json_expected
def replace_exercise(id_: int) -> ResponseReturnValue:
    try:
        exercise = (
            db.session.execute(
                select(Exercise)
                .where(Exercise.id == id_)
                .where(Exercise.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        exercise.name = data["name"]
    except (KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(exercise)),
        HTTPStatus.OK,
    )


@bp.route("/exercises/<int:id_>", methods=["DELETE"])
@session_required
def delete_exercise(id_: int) -> ResponseReturnValue:
    try:
        exercise = (
            db.session.execute(
                select(Exercise)
                .where(Exercise.id == id_)
                .where(Exercise.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(exercise)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/routines")
@session_required
def read_routines() -> ResponseReturnValue:
    routines = (
        db.session.execute(
            select(Routine)
            .where(Routine.user_id == session["user_id"])
            .options(selectinload(Routine.sections))
        )
        .scalars()
        .all()
    )
    return jsonify([to_dict(r) for r in routines])


@bp.route("/routines", methods=["POST"])
@session_required
@json_expected
def create_routine() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        routine = Routine(
            user_id=session["user_id"],
            name=data["name"],
            notes=data["notes"],
            sections=to_routine_sections(data["sections"]),
        )
    except (DeserializationError, KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(routine)

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(routine)),
        HTTPStatus.CREATED,
        {"Location": f"/routines/{routine.id}"},
    )


@bp.route("/routines/<int:id_>", methods=["PUT", "PATCH"])
@session_required
@json_expected
def update_routine(id_: int) -> ResponseReturnValue:
    try:
        routine = (
            db.session.execute(
                select(Routine)
                .where(Routine.id == id_)
                .where(Routine.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    try:
        if "name" in data or request.method == "PUT":
            routine.name = data["name"]
        if "notes" in data or request.method == "PUT":
            routine.notes = data["notes"]
        if "sections" in data or request.method == "PUT":
            routine.sections = to_routine_sections(data["sections"])
    except (DeserializationError, KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    try:
        db.session.commit()
    except IntegrityError as e:
        return jsonify({"details": str(e)}), HTTPStatus.CONFLICT

    return (
        jsonify(to_dict(routine)),
        HTTPStatus.OK,
    )


@bp.route("/routines/<int:id_>", methods=["DELETE"])
@session_required
def delete_routine(id_: int) -> ResponseReturnValue:
    try:
        routine = (
            db.session.execute(
                select(Routine)
                .where(Routine.id == id_)
                .where(Routine.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(routine)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT


@bp.route("/workouts")
@session_required
def read_workouts() -> ResponseReturnValue:
    workouts = (
        db.session.execute(
            select(Workout)
            .where(Workout.user_id == session["user_id"])
            .options(selectinload(Workout.elements))
        )
        .scalars()
        .all()
    )
    return jsonify([to_dict(w) for w in workouts])


@bp.route("/workouts", methods=["POST"])
@session_required
@json_expected
def create_workout() -> ResponseReturnValue:
    data = request.json

    assert isinstance(data, dict)

    try:
        routine = (
            (
                db.session.execute(
                    select(Routine)
                    .where(Routine.user_id == session["user_id"])
                    .where(Routine.id == data["routine_id"])
                )
                .scalars()
                .one()
            )
            if isinstance(data["routine_id"], int)
            else None
        )

        workout = Workout(
            user_id=session["user_id"],
            routine=routine,
            date=date.fromisoformat(data["date"]),
            notes=data["notes"],
            elements=to_workout_elements(data["elements"]),
        )
    except (DeserializationError, NoResultFound, KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.add(workout)

    db.session.commit()

    return (
        jsonify(to_dict(workout)),
        HTTPStatus.CREATED,
        {"Location": f"/workouts/{workout.id}"},
    )


@bp.route("/workouts/<int:id_>", methods=["PUT", "PATCH"])
@session_required
@json_expected
def update_workout(id_: int) -> ResponseReturnValue:
    try:
        workout = (
            db.session.execute(
                select(Workout)
                .where(Workout.id == id_)
                .where(Workout.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    data = request.json

    assert isinstance(data, dict)

    if "elements" in data or request.method == "PUT":
        for e in workout.elements:
            db.session.delete(e)

        db.session.flush()

    try:
        if "date" in data or request.method == "PUT":
            workout.date = date.fromisoformat(data["date"])
        if "notes" in data or request.method == "PUT":
            workout.notes = data["notes"]
        if "elements" in data or request.method == "PUT":
            workout.elements = to_workout_elements(data["elements"])
    except (DeserializationError, KeyError, ValueError) as e:
        return jsonify({"details": str(e)}), HTTPStatus.BAD_REQUEST

    db.session.commit()

    return (
        jsonify(to_dict(workout)),
        HTTPStatus.OK,
    )


@bp.route("/workouts/<int:id_>", methods=["DELETE"])
@session_required
def delete_workout(id_: int) -> ResponseReturnValue:
    try:
        workout = (
            db.session.execute(
                select(Workout)
                .where(Workout.id == id_)
                .where(Workout.user_id == session["user_id"])
            )
            .scalars()
            .one()
        )
    except (NoResultFound, ValueError):
        return "", HTTPStatus.NOT_FOUND

    db.session.delete(workout)
    db.session.commit()

    return "", HTTPStatus.NO_CONTENT



"""
Add reps, weight and rpe to routine_activity.

Revision ID: 22f3ddb25741
Revises: 8a0dc258bf2a
Create Date: 2023-04-01

"""

from typing import Union

import sqlalchemy as sa
from alembic import op

revision = "22f3ddb25741"
down_revision = "8a0dc258bf2a"
branch_labels = None
depends_on = None


check_constraints: list[tuple[str, Union[str, sa.ColumnElement[bool]]]] = [
    ("reps_type_integer", "typeof(reps) = 'integer'"),
    ("weight_type_real", "typeof(weight) = 'real'"),
    ("rpe_type_real", "typeof(rpe) = 'real'"),
    ("reps_ge_0", sa.column("reps") >= 0),
    ("weight_ge_0", sa.column("weight") >= 0),
    ("rpe_ge_0", sa.column("rpe") >= 0),
    ("rpe_le_10", sa.column("rpe") <= 10),
]


def upgrade() -> None:
    with op.batch_alter_table("routine_activity", schema=None) as batch_op:
        batch_op.add_column(sa.Column("reps", sa.Integer(), nullable=False, default=0))
        batch_op.add_column(sa.Column("weight", sa.Float(), nullable=False, default=0))
        batch_op.add_column(sa.Column("rpe", sa.Float(), nullable=False, default=0))
        for constraint_name, condition in check_constraints:
            batch_op.create_check_constraint(constraint_name, condition)


def downgrade() -> None:
    with op.batch_alter_table("routine_activity", schema=None) as batch_op:
        batch_op.drop_column("rpe")
        batch_op.drop_column("weight")
        batch_op.drop_column("reps")
        for constraint_name, _ in check_constraints:
            batch_op.drop_constraint(constraint_name, type_="check")



from flask import Blueprint, current_app, redirect, render_template, send_from_directory
from flask.typing import ResponseReturnValue

bp = Blueprint("assets", __name__, template_folder="templates")


def public_url() -> str:
    return current_app.config.get("PUBLIC_URL", "")


@bp.route("/")
def root() -> ResponseReturnValue:
    return redirect("app", code=301)


@bp.route("/app")
def app() -> ResponseReturnValue:
    return render_template("frontend.html", public_url=public_url())


@bp.route("/manifest.json")
def manifest() -> ResponseReturnValue:
    return render_template("manifest.json", public_url=public_url())


@bp.route("/<path:name>")
def frontend(name: str) -> ResponseReturnValue:
    return send_from_directory("frontend", name)



# ruff: noqa: T201

import argparse
import os
import sys
from pathlib import Path
from tempfile import NamedTemporaryFile

from valens import app, config, database as db, demo, version


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--version", action="version", version=version.get())

    subparsers = parser.add_subparsers(dest="subcommand")

    parser_config = subparsers.add_parser("config", help="create config")
    parser_config.set_defaults(func=create_config)
    parser_config.add_argument(
        "-d",
        dest="directory",
        type=Path,
        default=Path(),
        help="target directory for the to be created config file",
    )

    parser_upgrade = subparsers.add_parser("upgrade", help="upgrade database")
    parser_upgrade.set_defaults(func=upgrade)

    parser_run = subparsers.add_parser("run", help="run app on local development server")
    parser_run.set_defaults(func=run)
    parser_run.add_argument(
        "--public",
        action="store_true",
        help="make the server publicly available (sould be only used on a trusted network)",
    )
    parser_run.add_argument(
        "--port",
        metavar="NUMBER",
        type=int,
        default=5000,
        help="port to bind to",
    )

    parser_demo = subparsers.add_parser(
        "demo", help="run app with random example data (all changes are non-persistent)"
    )
    parser_demo.set_defaults(func=run_demo)
    parser_demo.add_argument(
        "--public",
        action="store_true",
        help="make the server publicly available (sould be only used on a trusted network)",
    )
    parser_demo.add_argument(
        "--port",
        metavar="NUMBER",
        type=int,
        default=5000,
        help="port to bind to",
    )

    args = parser.parse_args(sys.argv[1:])

    if not args.subcommand:
        parser.print_usage()
        return 2

    args.func(args)

    return 0


def create_config(args: argparse.Namespace) -> None:
    config_file = config.create_config_file(
        args.directory, Path.home() / ".local/share/valens/valens.db"
    )
    print(f"Created {config_file}")


def upgrade(_: argparse.Namespace) -> None:
    with app.app_context():
        config.check_config_file(os.environ.copy())
        db.upgrade()


def run(args: argparse.Namespace) -> None:
    with app.app_context():
        config.check_config_file(os.environ.copy())
        app.run("0.0.0.0" if args.public else "127.0.0.1", args.port)


def run_demo(args: argparse.Namespace) -> None:
    with NamedTemporaryFile() as f:
        demo.run(f"sqlite:///{f.name}", "0.0.0.0" if args.public else "127.0.0.1", args.port)



"""
Add type checking.

Revision ID: 4b6051594962
Revises: 3e7317eb4efa
Create Date: 2022-01-22

"""

from alembic import op

revision = "4b6051594962"
down_revision = "3e7317eb4efa"
branch_labels = None
depends_on = None

CHECK_CONSTRAINTS = [
    (
        "weight_type_real",
        "body_weight",
        "typeof(weight) = 'real'",
    ),
    (
        "chest_type_integer_or_null",
        "body_fat",
        "typeof(chest) = 'integer' or typeof(chest) = 'null'",
    ),
    (
        "abdominal_type_integer_or_null",
        "body_fat",
        "typeof(abdominal) = 'integer' or typeof(abdominal) = 'null'",
    ),
    (
        "tigh_type_integer_or_null",
        "body_fat",
        "typeof(tigh) = 'integer' or typeof(tigh) = 'null'",
    ),
    (
        "tricep_type_integer_or_null",
        "body_fat",
        "typeof(tricep) = 'integer' or typeof(tricep) = 'null'",
    ),
    (
        "subscapular_type_integer_or_null",
        "body_fat",
        "typeof(subscapular) = 'integer' or typeof(subscapular) = 'null'",
    ),
    (
        "suprailiac_type_integer_or_null",
        "body_fat",
        "typeof(suprailiac) = 'integer' or typeof(suprailiac) = 'null'",
    ),
    (
        "midaxillary_type_integer_or_null",
        "body_fat",
        "typeof(midaxillary) = 'integer' or typeof(midaxillary) = 'null'",
    ),
    (
        "intensity_type_integer",
        "period",
        "typeof(intensity) = 'integer'",
    ),
    (
        "position_type_integer",
        "routine_exercise",
        "typeof(position) = 'integer'",
    ),
    (
        "sets_type_integer",
        "routine_exercise",
        "typeof(sets) = 'integer'",
    ),
    (
        "position_type_integer",
        "workout_set",
        "typeof(position) = 'integer'",
    ),
    (
        "reps_type_integer_or_null",
        "workout_set",
        "typeof(reps) = 'integer' or typeof(reps) = 'null'",
    ),
    (
        "time_type_integer_or_null",
        "workout_set",
        "typeof(time) = 'integer' or typeof(time) = 'null'",
    ),
    (
        "weight_type_real_or_null",
        "workout_set",
        "typeof(weight) = 'real' or typeof(weight) = 'null'",
    ),
    (
        "rpe_type_real_or_null",
        "workout_set",
        "typeof(rpe) = 'real' or typeof(rpe) = 'null'",
    ),
]


def upgrade() -> None:
    for constraint_name, table_name, condition in CHECK_CONSTRAINTS:
        with op.batch_alter_table(table_name) as batch_op:
            batch_op.create_check_constraint(constraint_name, condition)


def downgrade() -> None:
    for constraint_name, table_name, _ in CHECK_CONSTRAINTS:
        with op.batch_alter_table(table_name) as batch_op:
            batch_op.drop_constraint(f"ck_{table_name}_{constraint_name}")



"""
Add foreign key ON DELETE cascade.

Revision ID: 3e7317eb4efa
Revises: 4cacd61cb0c5
Create Date: 2021-07-18

"""

from typing import Any

from alembic import op
from sqlalchemy import CheckConstraint, column

revision = "3e7317eb4efa"
down_revision = "4cacd61cb0c5"
branch_labels = None
depends_on = None


FOREIGN_KEY_CONSTRAINTS: list[tuple[str, str, str, str, tuple[Any, ...]]] = [  # type: ignore[misc]
    (
        "fk_body_weight_user_id_user",
        "body_weight",
        "user",
        "user_id",
        (CheckConstraint(column("weight") > 0, name="ck_body_weight_weight_gt_0"),),
    ),
    (
        "fk_body_fat_user_id_user",
        "body_fat",
        "user",
        "user_id",
        (
            CheckConstraint(column("chest") > 0, name="ck_body_fat_chest_gt_0"),
            CheckConstraint(column("abdominal") > 0, name="ck_body_fat_abdominal_gt_0"),
            CheckConstraint(column("tigh") > 0, name="ck_body_fat_tigh_gt_0"),
            CheckConstraint(column("tricep") > 0, name="ck_body_fat_tricep_gt_0"),
            CheckConstraint(column("subscapular") > 0, name="ck_body_fat_subscapular_gt_0"),
            CheckConstraint(column("suprailiac") > 0, name="ck_body_fat_suprailiac_gt_0"),
            CheckConstraint(column("midaxillary") > 0, name="ck_body_fat_midaxillary_gt_0"),
        ),
    ),
    (
        "fk_period_user_id_user",
        "period",
        "user",
        "user_id",
        (
            CheckConstraint(column("intensity") >= 1, name="ck_period_intensity_ge_1"),
            CheckConstraint(column("intensity") <= 4, name="ck_period_intensity_le_4"),
        ),
    ),
    (
        "fk_routine_exercise_routine_id_routine",
        "routine_exercise",
        "routine",
        "routine_id",
        (
            CheckConstraint(column("position") > 0, name="ck_routine_exercise_position_gt_0"),
            CheckConstraint(column("sets") > 0, name="ck_routine_exercise_sets_gt_0"),
        ),
    ),
    (
        "fk_routine_exercise_exercise_id_exercise",
        "routine_exercise",
        "exercise",
        "exercise_id",
        (
            CheckConstraint(column("position") > 0, name="ck_routine_exercise_position_gt_0"),
            CheckConstraint(column("sets") > 0, name="ck_routine_exercise_sets_gt_0"),
        ),
    ),
    (
        "fk_workout_set_workout_id_workout",
        "workout_set",
        "workout",
        "workout_id",
        (
            CheckConstraint(column("position") > 0, name="ck_workout_set_position_gt_0"),
            CheckConstraint(column("reps") > 0, name="ck_workout_set_reps_gt_0"),
            CheckConstraint(column("time") > 0, name="ck_workout_set_time_gt_0"),
            CheckConstraint(column("weight") > 0, name="ck_workout_set_weight_gt_0"),
            CheckConstraint(column("rpe") >= 0, name="ck_workout_set_rpe_ge_0"),
            CheckConstraint(column("rpe") <= 10, name="ck_workout_set_rpe_le_10"),
        ),
    ),
    (
        "fk_workout_set_exercise_id_exercise",
        "workout_set",
        "exercise",
        "exercise_id",
        (
            CheckConstraint(column("position") > 0, name="ck_workout_set_position_gt_0"),
            CheckConstraint(column("reps") > 0, name="ck_workout_set_reps_gt_0"),
            CheckConstraint(column("time") > 0, name="ck_workout_set_time_gt_0"),
            CheckConstraint(column("weight") > 0, name="ck_workout_set_weight_gt_0"),
            CheckConstraint(column("rpe") >= 0, name="ck_workout_set_rpe_ge_0"),
            CheckConstraint(column("rpe") <= 10, name="ck_workout_set_rpe_le_10"),
        ),
    ),
    ("fk_exercise_user_id_user", "exercise", "user", "user_id", ()),
    ("fk_workout_user_id_user", "workout", "user", "user_id", ()),
    ("fk_workout_routine_id_routine", "workout", "routine", "routine_id", ()),
    ("fk_routine_user_id_user", "routine", "user", "user_id", ()),
]


def upgrade() -> None:
    for (
        constraint_name,
        source_table,
        referent_table,
        local_col,
        table_args,
    ) in FOREIGN_KEY_CONSTRAINTS:
        with op.batch_alter_table(source_table, schema=None, table_args=table_args) as batch_op:
            batch_op.drop_constraint(constraint_name, type_="foreignkey")
            batch_op.create_foreign_key(
                batch_op.f(constraint_name), referent_table, [local_col], ["id"], ondelete="CASCADE"
            )


def downgrade() -> None:
    for (
        constraint_name,
        source_table,
        referent_table,
        local_col,
        table_args,
    ) in FOREIGN_KEY_CONSTRAINTS:
        with op.batch_alter_table(source_table, schema=None, table_args=table_args) as batch_op:
            batch_op.drop_constraint(constraint_name, type_="foreignkey")
            batch_op.create_foreign_key(
                batch_op.f(constraint_name), referent_table, [local_col], ["id"]
            )



# ruff: noqa: T201

import sqlite3
from datetime import datetime
from pathlib import Path
from shutil import copy
from time import sleep

from alembic import command, runtime, script
from alembic.config import Config
from flask import current_app, g
from sqlalchemy import Connection, Engine, create_engine, event, inspect, pool
from sqlalchemy.orm import Session, scoped_session, sessionmaker
from werkzeug.local import LocalProxy

from valens import config, models

alembic_cfg = Config()
alembic_cfg.set_main_option("script_location", "valens:migrations")


def db_file() -> Path:
    return Path(current_app.config["DATABASE"].removeprefix("sqlite:///"))


def db_dir() -> Path:
    return db_file().parent


def upgrade_lock_file() -> Path:
    return db_dir() / "valens_upgrade.lock"


def get_engine() -> Engine:
    config.check_app_config()
    db_dir().mkdir(exist_ok=True)
    return create_engine(current_app.config["DATABASE"])


def get_scoped_session() -> scoped_session[Session]:
    return scoped_session(
        sessionmaker(autocommit=False, autoflush=False, bind=get_engine(), future=True)
    )


def get_session() -> Session:
    if "db_session" not in g:
        if not inspect(get_engine()).get_table_names():
            init()
        g.db_session = get_scoped_session()()

    _upgrade(g.db_session.connection())

    return g.db_session


session: Session = LocalProxy(get_session)  # type: ignore[assignment]


def remove_session() -> None:
    get_scoped_session().remove()


def init() -> None:
    print("Creating database")

    models.Base.query = get_scoped_session().query_property()
    models.Base.metadata.create_all(bind=get_engine())

    command.stamp(alembic_cfg, "head")


def upgrade() -> None:
    get_session()


def _upgrade(connection: Connection) -> None:
    current = runtime.migration.MigrationContext.configure(connection).get_current_revision()
    head = script.ScriptDirectory.from_config(alembic_cfg).get_current_head()

    if current != head:
        try:
            upgrade_lock_file().touch(exist_ok=False)

            print(f"Upgrading database from {current} to {head}")

            copy(
                db_file(),
                db_file().with_suffix(
                    f".db.backup_{current}_{datetime.now().isoformat(timespec='seconds')}"
                ),
            )
            command.upgrade(alembic_cfg, "head")

            upgrade_lock_file().unlink()

        except FileExistsError:
            print("Waiting for completion of database upgrade")

            while upgrade_lock_file().exists():
                sleep(1)

        except Exception as e:
            print(f"Database upgrade failed: {e}")

            upgrade_lock_file().unlink()


@event.listens_for(Engine, "connect")
def _set_sqlite_pragma(
    dbapi_connection: sqlite3.Connection, _: pool.base._ConnectionRecord
) -> None:
    if current_app.config["SQLITE_FOREIGN_KEY_SUPPORT"]:
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()



from flask import Flask

from . import api, assets

app = Flask(__name__)

app.config.from_object("valens.default_config")
app.config.from_envvar("VALENS_CONFIG", silent=True)

app.jinja_env.lstrip_blocks = True
app.jinja_env.trim_blocks = True

app.register_blueprint(assets.bp)
app.register_blueprint(api.bp)



from __future__ import annotations

import datetime
import random
from dataclasses import dataclass

from valens import app, database as db
from valens.models import (
    BodyFat,
    BodyWeight,
    Exercise,
    Period,
    Routine,
    RoutineActivity,
    RoutineSection,
    Sex,
    User,
    Workout,
    WorkoutRest,
    WorkoutSet,
)


def run(database: str, host: str = "127.0.0.1", port: int = 5000) -> None:
    app.config["DATABASE"] = database
    app.config["SECRET_KEY"] = b"TEST_KEY"
    with app.app_context():
        for user in users():
            db.session.add(user)
        db.session.commit()
        app.run(host, port)


def users() -> list[User]:
    result = []
    for user_id, name, sex in [(1, "Alice", Sex.FEMALE), (2, "Bob", Sex.MALE)]:
        exercises, routines, workouts = _workouts(user_id)
        body_weight = _body_weight(user_id)
        result.append(
            User(
                id=user_id,
                name=name,
                sex=sex,
                body_weight=body_weight,
                body_fat=_body_fat(body_weight, user_id),
                period=_period(user_id),
                exercises=exercises,
                routines=routines,
                workouts=workouts,
            )
        )
    return result


def _body_weight(user_id: int = 1) -> list[BodyWeight]:
    day = datetime.date.today()
    weight = random.uniform(50, 100)
    values = [(day, weight)]

    for i in range(1, 365):
        weight += random.gauss(-0.1 if i % 2 == 0 else 0.2, 0.2)
        if random.randint(0, 2) == 0:
            continue
        values.append((datetime.date.today() - datetime.timedelta(days=i), weight))

    return [BodyWeight(user_id=user_id, date=d, weight=w) for d, w in values]


def _body_fat(body_weight: list[BodyWeight], user_id: int = 1) -> list[BodyFat]:
    initial_bw = body_weight[-1].weight
    initial_bf: tuple[int, ...] = (
        random.randint(5, 20),
        random.randint(10, 30),
        random.randint(10, 30),
        random.randint(10, 30),
        random.randint(5, 20),
        random.randint(5, 20),
        random.randint(5, 20),
    )
    bf = []
    i = 0

    while i < len(body_weight):
        date = body_weight[i].date
        factor = body_weight[i].weight / initial_bw
        bf.append((date, tuple(int(e * factor) + random.randint(0, 1) for e in initial_bf)))
        i += 7

    return [
        BodyFat(
            user_id=user_id,
            date=date,
            chest=che,
            abdominal=abd,
            tigh=tig,
            tricep=tri,
            subscapular=sub,
            suprailiac=sup,
            midaxillary=mid,
        )
        for date, (che, abd, tig, tri, sub, sup, mid) in bf
    ]


def _period(user_id: int = 1) -> list[Period]:
    day = datetime.date.today() - datetime.timedelta(days=random.randint(7, 33))
    values = []

    for _ in range(13):
        previous = 4
        for d in range(7):
            intensity = random.randint(max(0, 3 - d), min(4, previous))
            previous = intensity

            if intensity == 0:
                continue

            values.append((day + datetime.timedelta(days=d), intensity))

        day -= datetime.timedelta(days=28 + int(random.gauss(0, 3)))

    return [Period(user_id=user_id, date=d, intensity=i) for d, i in values]


@dataclass
class ExerciseType:
    reps: bool
    time: bool
    weight: bool
    rpe: bool


def _workouts(user_id: int = 1) -> tuple[list[Exercise], list[Routine], list[Workout]]:
    exercise_names = {
        "Barbell Back Squat": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Barbell Bench Press": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Barbell Deadlift": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Barbell Lunge": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Barbell Romanian Deadlift": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Barbell Row": ExerciseType(reps=True, time=True, weight=True, rpe=True),
        "Dip": ExerciseType(reps=True, time=True, weight=False, rpe=True),
        "Dumbbell Curl": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Dumbbell Press": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Dumbbell Shoulder Press": ExerciseType(reps=True, time=False, weight=True, rpe=False),
        "Hip Thrust": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Lat Pulldown": ExerciseType(reps=True, time=True, weight=True, rpe=True),
        "Plank": ExerciseType(reps=False, time=True, weight=False, rpe=False),
        "Pull Up": ExerciseType(reps=True, time=True, weight=False, rpe=True),
        "Push Up": ExerciseType(reps=True, time=True, weight=False, rpe=True),
        "Seated Leg Curl": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Seated Leg Extension": ExerciseType(reps=True, time=False, weight=True, rpe=True),
        "Walking Lunge": ExerciseType(reps=True, time=False, weight=True, rpe=False),
    }
    exercises = [
        Exercise(user_id=user_id, name=name)
        for name in random.sample(sorted(exercise_names.keys()), k=len(exercise_names))
    ]

    routine_names = ["A", "B", "C", "D"]
    routines = [
        Routine(
            id=(user_id - 1) * len(routine_names) + i,
            user_id=user_id,
            name=f"Training {t}",
            sections=[
                RoutineSection(
                    position=p,
                    rounds=random.randint(2, 5),
                    parts=[
                        RoutineActivity(
                            position=1,
                            exercise=e,
                            reps=0,
                            time=0,
                            weight=0.0,
                            rpe=0.0,
                            automatic=False,
                        ),
                        RoutineActivity(
                            position=2,
                            reps=0,
                            time=60,
                            weight=0.0,
                            rpe=0.0,
                            automatic=True,
                        ),
                    ],
                )
                for p, e in enumerate(random.sample(exercises, random.randint(5, 8)), start=1)
            ],
        )
        for i, t in enumerate(routine_names, start=1)
    ]

    workouts = [
        Workout(
            user_id=user_id,
            date=datetime.date.today()
            - datetime.timedelta(days=len(routines) * 13 * 7)
            + datetime.timedelta(days=(quarter * 13 * 7) + (week * 7) + day),
            elements=[
                element
                for position, (exercise, reps, time, weight, rpe) in enumerate(
                    [
                        (routine_activity.exercise, reps, time, weight, rpe)
                        for routine_section in routines[quarter].sections
                        for routine_activity in routine_section.parts
                        if isinstance(routine_activity, RoutineActivity)
                        and routine_activity.exercise is not None
                        for reps, time, weight, rpe in (
                            [
                                (
                                    (
                                        5 + week + random.randint(0, 1)
                                        if exercise_names[routine_activity.exercise.name].reps
                                        else None
                                    ),
                                    (
                                        (
                                            random.randint(3, 4)
                                            if exercise_names[routine_activity.exercise.name].reps
                                            else 10 + 5 * week + 5 * random.randint(0, 2)
                                        )
                                        if exercise_names[routine_activity.exercise.name].time
                                        else None
                                    ),
                                    (
                                        5 + (week + random.randint(0, 1)) * 0.5
                                        if exercise_names[routine_activity.exercise.name].weight
                                        else None
                                    ),
                                    (
                                        min(7 + (week % 4) + (random.randint(0, 1) * 0.5), 10)
                                        if exercise_names[routine_activity.exercise.name].rpe
                                        else None
                                    ),
                                )
                            ]
                        )
                        for _ in range(routine_section.rounds)
                    ],
                    start=0,
                )
                for element in [
                    WorkoutSet(
                        position=2 * position + 1,
                        exercise=exercise,
                        reps=reps,
                        time=time,
                        weight=weight,
                        rpe=rpe,
                    ),
                    WorkoutRest(
                        position=2 * position + 2,
                        target_time=60,
                    ),
                ]
            ],
            routine=routines[quarter],
        )
        for quarter in range(len(routines))
        for week in range(13)
        for day in [0, 3]
    ]
    return (exercises, routines, workouts)
